from .aggregation import majority
from .chain_of_thought import ChainOfThought
from .chain_of_thought_with_hint import ChainOfThoughtWithHint
from .knn import KNN
from .multi_chain_comparison import MultiChainComparison
from .predict import Predict
from .program_of_thought import ProgramOfThought
from .react import ReAct
from .retry import Retry
from .parallel import Parallel